
class Scalar:
    def __init__(self, value=0.0):
        self.value = value

    def __repr__(self):
        return f"Scalar({self.value})"

    def __str__(self):
        return f"Scalar({self.value})"
